#
# Abstraction of a set of message classes
#

import os
import glob
import ConfigParser
import logging

logger = logging.getLogger('msg.conf')
    
def load_message_classes(dir):
    """Load a set of message classes from a configuration directory and return a messageclass set"""

    if not os.path.isdir(dir):
        raise os.error("Directory does not exist : %s"%dir)
    parser = ConfigParser.SafeConfigParser()
    parser.read(glob.glob('%s/*.conf'%dir))

    mcs = {}
    for s in parser.sections():
        mc = create_messageclass(s, parser.items(s))
        mcs[mc.name]=mc
    logger.info("Have message classes : %s"%mcs.keys())
    return mcs

def create_messageclass(name, fields):
    """Create a message class out of a list of (key, value) tuples.  Allowed keys are:
       required, headers, destination, spec_version

       >>> mc = create_messageclass('foo', {'required': 'foo', 'headers' : 'foo bar', 'spec_version' : '0.91'})
       >>> mc.required = ['foo']
       >>> mc.headers = ['foo', 'bar']
       >>> mc.spec_version = '0.91'
       """
    m = dict(fields)
    for key in ['required', 'headers']:
        if m.has_key(key):
            m[key] = m[key].lower().split(' ')
    return MessageClass(name=name,
                        required = m.get('required',None),
                        headers = m.get('headers', None),
                        destination = m.get('destination', None),
                        spec_version = m.get('spec_version', None))
                        
    

class MessageClass:
    def __init__(self, name, required = None, headers = None,
                 destination = None, spec_version = None):
        """       Multi-valued fileds are represented by space separated list"""
        self.name = name
        self.required = required
        self.headers = headers
        self.destination = destination        
    
    def validate(self, record):
        """
        >>> mc = MessageClass('foo', required = ['x', 'y'], headers = ['x'], destination = 'foobar')

        # A valid record
        >>> record = {'x':'1', 'y':'2', 'z':3}
        >>> mc.validate(record)
        True

        # and now one without all required fields
        >>> record = {'x':'1'}
        >>> mc.validate(record)
        False

        """
        for ra in self.required:
            if not record.has_key(ra):
                logger.warn('Incomplete record, missing required attribute: %s'%ra)
                logger.info("Required Fields are : %s"%self.required)
                return False
        return True

    def convertToMessages(self, records, bulk) :
        """ Given a set of records, convert into a set of messages ready to be sent
        the to message broker.  If bulk == True, we try and bulk the records into
        groups, where all records in a single message have the same value for all
        headers"""
        if bulk:
            recordGroups = _groupby(records, lambda x: tuple(map(x.get, self.headers)))
            return [self.composeMessage(group) for  group in recordGroups]

        return [self.composeMessage([record]) for record in records]

    def composeMessage(self, records):
        """Compose a message according to this message class, consisting of these set
        of records to send.

        Assumes that all records have the same headers"""

        body = ''.join([r.dumps() for r in records])
        msg = {'destination': self.destination, 'body' : body}
        if self.headers:
            msg['header'] = dict([(k, records[0].get(k)) for k in self.headers])
            logger.info("Message is : %s"%msg)
        return msg

class _groupby(dict):
    def __init__(self, seq, key=lambda x:x):
        for value in seq:
            k = key(value)
            self.setdefault(k, []).append(value)
    __iter__ = dict.itervalues


if __name__ == "__main__":
    import doctest
    doctest.testmod()
